// vim: set foldmethod=marker :

#include <avr/io.h>
#include <stdint.h>

#define NUM_SENSORS 6

// 1 - Left sensor is connected to PA7
// 0 - Right sensor is connected to PA0
#define SENSORS_ALIGN 1

// 0 - black line on white background
// 1 - white line on black background
#define WHITE_LINE 0

#define SENSORS_MASK ((uint8_t)(0xFF << (8 - NUM_SENSORS)))

#if SENSORS_ALIGN
#	define SENSORS_SHIFT(val) { val >>= 1;}
#	define SENSORS_MASK_LEFT 0x80
#else
#	define SENSORS_SHIFT(val) { val <<= 1;}
#	define SENSORS_MASK_LEFT 0x01
#endif

#define SENSORS_MAX_TIME 10000
#define SENSORS_DISCHARGE_TIME 200

static uint16_t calibratedMin[NUM_SENSORS];
static uint16_t calibratedMax[NUM_SENSORS];

void read_sensors_raw(uint16_t* values)/*{{{*/
{
	uint8_t start;
	uint8_t delta;
	uint8_t last;
	uint8_t sample;
	uint8_t i;
	uint16_t time;

	time = 0;

	TCCR2 = _BV(CS20);
	DDRA |= SENSORS_MASK;
	PORTA &= ~SENSORS_MASK;
	start = TCNT2;
	do {
		delta = TCNT2 - start;
	} while (delta < SENSORS_DISCHARGE_TIME);
	
	DDRA &= ~SENSORS_MASK;
	last = PINA & SENSORS_MASK;
	start =  TCNT2;
	while (time < SENSORS_MAX_TIME) {
		uint8_t current = TCNT2;
		delta = start - current;
		start = current;
		time += delta;

		sample = PINA & SENSORS_MASK;
		if (last == sample) {
			continue;
		}
		last = sample;
		current = SENSORS_MASK_LEFT;
		for (i =0; i < NUM_SENSORS; i++) {
			if (sample & current) {
				values[i] = time;
			}
			SENSORS_SHIFT(current);
		}
	}
}/*}}}*/

void calibrate(void)/*{{{*/
{
	uint8_t i;
	uint16_t values[8];
	read_sensors_raw(values);
	for (i = 0; i < NUM_SENSORS; i++) {
		if (!calibratedMin[i] || calibratedMin[i] > values[i]) {
			calibratedMin[i] = values[i];
		}
		if (calibratedMax[i] < values[i]) {
			calibratedMax[i] = values[i];
		}
	}
}/*}}}*/

void read_sensors_calibrated(uint16_t* values)/*{{{*/
{
	uint8_t i;
	read_sensors_raw(values);
	for (i = 0; i < NUM_SENSORS; i++) {
		uint16_t value = values[i];
		uint16_t calMax = calibratedMax[i];
		if (!calMax) {
			calMax = SENSORS_MAX_TIME;
		}
		if (value < calibratedMin[i]) {
			value = calibratedMin[value];
		}
		value = value - calibratedMin[i];
		values[i] = (uint32_t)value * SENSORS_MAX_TIME / (calMax - calibratedMin[i]);
	}
}/*}}}*/

